from threading import Lock, Semaphore
from typing import Callable, Generic, TypeVar

V = TypeVar("V")


class ValueContext(Generic[V]):
    def __init__(
        self,
        value: V | None,
        on_exit: Callable[[], None] | None,
        on_set_value: Callable[[V], None] | None,
    ) -> None:
        super().__init__()
        self._value: V | None = value
        self._on_exit: Callable[[], None] | None = on_exit
        self._on_set_value: Callable[[V], None] | None = on_set_value

    @property
    def value(self) -> V | None:
        return self._value

    def update_value(self, value: V) -> None:
        if self._on_set_value is not None:
            assert value is not None, "value must not be None"
            self._value = value
            self._on_set_value(value)

    def __enter__(self) -> "ValueContext[V]":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if self._on_exit is not None:
            self._on_exit()


class ValueSignal(Generic[V]):
    def __init__(self) -> None:
        super().__init__()
        self._value: V | None = None
        self._semaphore: Semaphore = Semaphore(1)
        self._lock: Lock = Lock()
        self._acquire_count: int = 0

    def context(self) -> ValueContext[V]:
        with self._lock:
            if self._value is None:
                self._acquire_count += 1
            else:
                return ValueContext(
                    value=self._value,
                    on_exit=None,
                    on_set_value=None,
                )
        self._semaphore.acquire()

        return ValueContext(
            value=self._value,
            on_exit=self._exit_context,
            on_set_value=self._set_value,
        )

    def _set_value(self, value: V) -> None:
        acquire_count: int
        with self._lock:
            self._value = value
            acquire_count = self._acquire_count

        for _ in range(acquire_count):
            self._semaphore.release()

    def _exit_context(self) -> None:
        with self._lock:
            if self._value is not None:
                return
            self._acquire_count -= 1
        self._semaphore.release()



from dataclasses import dataclass
from threading import Event, Lock
from typing import Callable, Iterable

from .errors import InterruptionError


@dataclass
class SegmentDescription:
    offset: int
    length: int
    completed_length: int


# thread safe
class Segment:
    def __init__(
        self,
        offset: int,
        length: int,
        completed_length: int,
        send_back: Callable[["Segment"], None],
    ) -> None:
        self._offset: int = offset
        self._length: int = length
        self._locked_length: int = completed_length
        self._completed_length: int = completed_length
        self._send_back: Callable[["Segment"], None] = send_back
        self._lock: Lock = Lock()
        self._interrupted: bool = False

    @property
    def lock(self) -> Lock:
        return self._lock

    @property
    def offset(self) -> int:
        return self._offset

    @property
    def length(self) -> int:
        return self._length

    @property
    def locked_length(self) -> int:
        return self._locked_length

    @property
    def completed_length(self) -> int:
        return self._completed_length

    @property
    def is_completed(self) -> bool:
        with self._lock:
            return self._completed_length >= self._length

    def lock_next_length(self, delta_length: int) -> int:
        with self._lock:
            if self._interrupted:
                raise InterruptionError()
            origin_locked_length = self._locked_length
            self._locked_length = min(origin_locked_length + delta_length, self._length)
            return self._locked_length - origin_locked_length

    def submit(self, delta_length: int) -> None:
        with self._lock:
            next_completed_length = self._completed_length + delta_length
            if next_completed_length > self._locked_length:
                raise ValueError("cannot submit more than locked length")
            self._completed_length = next_completed_length

    def complete(self) -> None:
        with self._lock:
            self._locked_length = self._length
            self._completed_length = self._length

    def interrupt(self) -> None:
        with self._lock:
            self._interrupted = True

    def dispose(self) -> None:
        with self._lock:
            self._send_back(self)


@dataclass
class _Node:
    taken: bool
    done: bool
    segment: Segment


# not thread safe
class Serial:
    def __init__(
        self,
        length: int,
        min_segment_length: int,
        descriptions: Iterable[SegmentDescription] | None = None,
    ) -> None:
        assert length > 0, "length must be greater than 0"
        assert min_segment_length > 1, "min_segment_length must be greater than 1"
        self._length: int = length
        self._min_segment_length: int = min_segment_length
        self._nodes_lock: Lock = Lock()
        self._nodes: list[_Node] = []
        self._did_dispose: bool = False
        self._wait_interrupted_segments: int = 0
        self._wait_all_interrupted: Event = Event()

        if descriptions is None:
            self._nodes.append(
                _Node(
                    taken=False,
                    done=False,
                    segment=Segment(
                        offset=0,
                        length=length,
                        completed_length=0,
                        send_back=self._receive_segment,
                    ),
                )
            )
        else:
            offset: int = 0
            for description in sorted(list(descriptions), key=lambda d: d.offset):
                if description.offset != offset:
                    raise ValueError(
                        f"except segment offset to be continuous: wanna {offset} but got {description.offset}"
                    )
                if description.completed_length > description.length:
                    raise ValueError(
                        f"segment completed length {description.completed_length} cannot be greater than its length {description.length}"
                    )
                self._nodes.append(
                    _Node(
                        taken=False,
                        done=description.completed_length >= description.length,
                        segment=Segment(
                            offset=description.offset,
                            length=description.length,
                            completed_length=description.completed_length,
                            send_back=self._receive_segment,
                        ),
                    )
                )
                offset += description.length
                if offset > length:
                    raise ValueError(f"segment's tail {offset} exceeds total length {length}")
            if offset != length:
                raise ValueError(f"end of segments {offset} does not match total length {length}")

    @property
    def length(self) -> int:
        return self._length

    @property
    def is_completed(self) -> bool:
        with self._nodes_lock:
            return all(node.done for node in self._nodes)

    def snapshot(self) -> list[SegmentDescription]:
        with self._nodes_lock:
            descriptions = [
                SegmentDescription(
                    offset=node.segment.offset,
                    length=node.segment.length,
                    completed_length=node.segment.completed_length,
                )
                for node in self._nodes
            ]
        descriptions.sort(key=lambda d: d.offset)

        return descriptions

    def interrupt(self) -> None:
        self._interrupt_taken_segments()

    def dispose(self) -> None:
        wait_count: int = self._interrupt_taken_segments()
        if wait_count > 0:
            self._wait_all_interrupted.wait()

    def _interrupt_taken_segments(self) -> int:
        wait_count: int = 0
        with self._nodes_lock:
            if self._did_dispose:
                return 0
            for node in self._nodes:
                if node.taken:
                    node.segment.interrupt()
                    wait_count += 1
            self._wait_interrupted_segments = wait_count
            self._wait_all_interrupted.clear()
            self._did_dispose = True
        return wait_count

    def pop_segment(self) -> Segment | None:
        if self._did_dispose:
            return None

        with self._nodes_lock:
            useable_sorted_keys: list[tuple[int, int, int]] = []
            for node_index, node in enumerate(self._nodes):
                if node.done:
                    continue
                taken_index: int = 0
                segment = node.segment
                remain_length = segment.length - segment.locked_length

                if node.taken:
                    taken_index = 1
                    if remain_length < 2 * self._min_segment_length:
                        # 低于此值，若切割它必然产生一个小于 min_segment_length 的片段
                        continue
                useable_sorted_keys.append((taken_index, -remain_length, node_index))

            for _, _, node_index in sorted(useable_sorted_keys):
                node = self._nodes[node_index]
                segment = node.segment
                if not node.taken:
                    node.taken = True
                    return segment

                with segment.lock:
                    cutted_segment = self._try_to_cut_segment(segment)
                    if cutted_segment:
                        return cutted_segment

            return None

    # all locks will be locked by the caller
    def _try_to_cut_segment(self, segment: Segment) -> Segment | None:
        remain_length = segment.length - segment.locked_length
        cutted_length = remain_length // 2

        if cutted_length < self._min_segment_length:
            # 两次上锁之间，状态可能发生变化，需要再校验一次
            return None

        segment._length -= cutted_length
        cutted_segment = Segment(
            offset=segment.offset + segment.length,
            length=cutted_length,
            completed_length=0,
            send_back=self._receive_segment,
        )
        self._nodes.append(
            _Node(
                taken=True,
                done=False,
                segment=cutted_segment,
            )
        )
        return cutted_segment

    def _receive_segment(self, segment: Segment) -> None:
        with self._nodes_lock:
            received_node: _Node | None = None
            for node in self._nodes:
                if node.segment == segment:
                    received_node = node
                    break

            if received_node is None:
                return
            received_node.taken = False
            if segment.completed_length >= segment.length:
                received_node.done = True

            if self._did_dispose:
                self._wait_interrupted_segments -= 1
                if self._wait_interrupted_segments <= 0:
                    self._wait_all_interrupted.set()



from collections.abc import Callable, Iterable, Iterator
from dataclasses import dataclass
from pathlib import Path
from threading import Event, Lock

from .common import HTTPOptions, Retry
from .file import CanRetryError, FileDownloader, InterruptionError
from .type import FileDownloadError, RetryError, Task, TaskError, TooManyRetriesError
from .utils import list_safe_remove


@dataclass
class _FileNode:
    task: Task
    downloader: FileDownloader
    ladder: int
    current_ladder_failure_count: int
    executors_count: int


# thread safe
class FilesGroup:
    def __init__(
        self,
        tasks_iter: Iterator[Task],
        window_width: int,
        failure_ladder: Iterable[int],
        min_segment_length: int,
        once_fetch_size: int,
        skip_existing: bool,
        timeout: float,
        retry: Retry,
        on_task_completed: Callable[[Task], None] | None,
        on_task_skipped: Callable[[Task], None] | None,
        on_task_failed: Callable[[TaskError], None] | None,
        on_task_failed_with_retry_error: Callable[[RetryError], None] | None,
    ) -> None:
        self._tasks_iter: Iterator[Task] = tasks_iter
        self._window_width: int = window_width
        self._min_segment_length: int = min_segment_length
        self._once_fetch_size: int = once_fetch_size
        self._skip_existing: bool = skip_existing
        self._timeout: float = timeout
        self._retry: Retry = retry
        self._on_task_completed: Callable[[Task], None] = on_task_completed or (lambda _: None)
        self._on_task_skipped: Callable[[Task], None] = on_task_skipped or (lambda _: None)
        self._on_task_failed: Callable[[TaskError], None] | None = on_task_failed
        self._on_task_failed_with_retry_error: Callable[[RetryError], None] = on_task_failed_with_retry_error or (
            lambda _: None
        )

        self._lock: Lock = Lock()
        self._did_call_dispose: bool = False
        self._failure_error: TaskError | None = None
        self._failure_ladder: tuple[int, ...] = tuple(failure_ladder)
        self._ladder_nodes: list[list[_FileNode]] = [[]] * len(self._failure_ladder)
        self._maybe_create_new_executors: Event = Event()

        assert len(self._failure_ladder) > 0, "failure ladder must not be empty"
        for ladder_failure_limit in self._failure_ladder:
            assert ladder_failure_limit > 0, "ladder failure limit must be greater than zero"

        self._maybe_create_new_executors.set()

    @property
    def is_empty(self) -> bool:
        with self._lock:
            for ladder_nodes in self._ladder_nodes:
                if ladder_nodes:
                    return False
            return True

    def raise_if_failure(self) -> None:
        with self._lock:
            if self._failure_error is not None:
                raise self._failure_error

    def dispose(self) -> None:
        downloaders: list[FileDownloader] = []
        with self._lock:
            if self._did_call_dispose:
                return
            self._did_call_dispose = True
            for ladder_nodes in self._ladder_nodes:
                for node in ladder_nodes:
                    downloaders.append(node.downloader)

        for downloader in downloaders:
            downloader.dispose()

    def wait_windows_update(self) -> None:
        # window 中的 node 被删除了，window 才有可能继续滑动，或者，有一个任务失败了，导致 node 被释放
        # 用户监听到该事件后，可以重新调用 pop_downloading_executor() 试图滑动 window 开始新任务
        self._maybe_create_new_executors.wait()

    def pop_downloading_executor(self) -> Callable[[], None] | None:
        with self._lock:
            if self._failure_error is not None:
                return None
            result = self._node_and_executor()
            if not result:
                self._maybe_create_new_executors.clear()
                return None

        # running in background thread
        def run_downloader_executor(node: _FileNode, executor: Callable[[], None]) -> None:
            try:
                executor()
            except InterruptionError:
                pass

            except CanRetryError as error:
                with self._lock:
                    success = self._increase_failure_count(node)
                    if success:
                        self._on_task_failed_with_retry_error(RetryError(node.task, error))
                        self._maybe_create_new_executors.set()
                    else:
                        self._remove_node(node)
                        self._emit_failure_error(TooManyRetriesError(node.task, error))

            except Exception as error:
                with self._lock:
                    self._remove_node(node)
                    self._emit_failure_error(FileDownloadError(node.task, error))

            finally:
                with self._lock:
                    node.executors_count -= 1
                    if node.executors_count <= 0:
                        completed_path = node.downloader.try_complete()
                        if completed_path is not None:
                            self._remove_node(node)
                            self._on_task_completed(node.task)

        node, executor = result
        node.executors_count += 1

        return lambda: run_downloader_executor(node, executor)

    def _increase_failure_count(self, node: _FileNode) -> bool:
        ladder_failure_limit = self._failure_ladder[node.ladder]
        origin_ladder_nodes = self._ladder_nodes[node.ladder]
        if node in origin_ladder_nodes:
            node.current_ladder_failure_count += 1
            if node.current_ladder_failure_count >= ladder_failure_limit:
                origin_ladder_nodes.remove(node)
                node.ladder += 1
                node.current_ladder_failure_count = 0
                if node.ladder >= len(self._ladder_nodes):
                    return False
                self._ladder_nodes[node.ladder].append(node)
        return True

    def _remove_node(self, node: _FileNode):
        if node.ladder < len(self._ladder_nodes):
            ladder_nodes = self._ladder_nodes[node.ladder]
            removed_node = list_safe_remove(ladder_nodes, node)
            if removed_node is not None:
                self._maybe_create_new_executors.set()

    def _node_and_executor(self) -> tuple[_FileNode, Callable[[], None]] | None:
        window_nodes = self._ladder_nodes[0]
        for node in window_nodes:
            executor = node.downloader.pop_downloading_task()
            if executor:
                return node, executor

        while len(window_nodes) < self._window_width:
            task = next(self._tasks_iter, None)
            if not task:
                break
            node = self._create_file_node(task)
            if node:
                window_nodes.append(node)
                executor = node.downloader.pop_downloading_task()
                if executor:
                    return node, executor

        for i in range(1, len(self._ladder_nodes)):
            ladder_nodes = self._ladder_nodes[i]
            for node in ladder_nodes:
                executor = node.downloader.pop_downloading_task()
                if executor:
                    return node, executor

        return None

    def _create_file_node(self, task: Task) -> _FileNode | None:
        task_file_path = Path(task.file)
        if task_file_path.exists():
            if self._skip_existing:
                self._on_task_skipped(task)
                return None
            if not task_file_path.is_file():
                raise ValueError(f"Task file path {task_file_path} exists but is not a file.")
            task_file_path.unlink()

        if callable(task.url):
            task.url = task.url()

        http_options = HTTPOptions(
            url=task.url,
            timeout=self._timeout,
            retry=self._retry,
            headers=task.headers,
            cookies=task.cookies,
        )
        try:
            downloader = FileDownloader(
                file_path=task_file_path,
                http_options=http_options,
                min_segment_length=self._min_segment_length,
                once_fetch_size=self._once_fetch_size,
            )
        except Exception as error:
            raise FileDownloadError(task, error) from error

        return _FileNode(
            task=task,
            downloader=downloader,
            ladder=0,
            current_ladder_failure_count=0,
            executors_count=0,
        )

    def _emit_failure_error(self, error: TaskError) -> None:
        if self._on_task_failed:
            self._on_task_failed(error)
        else:
            self._failure_error = error



import shutil
from enum import Enum
from pathlib import Path
from threading import Lock
from typing import Callable

import requests

from ..common import HTTPOptions
from ..retry import CAN_RETRY_STATUS_CODES, is_exception_can_retry
from .common import chunk_name
from .errors import CanRetryError, RangeDownloadFailedError
from .range_downloader import RangeDownloader
from .segment import Segment
from .utils import clean_path


class _SingletonPhase(Enum):
    NOT_STARTED = 0
    POPPED_TASK = 1
    COMPLETED = 2
    FAILED = 3


class FileDownloader:
    def __init__(
        self,
        file_path: Path,
        http_options: HTTPOptions,
        min_segment_length: int,
        once_fetch_size: int,
        excepted_etag: str | None = None,
    ) -> None:
        assert not file_path.exists(), "file already exists"

        self._file_path: Path = file_path
        self._http_options: HTTPOptions = http_options
        self._min_segment_length: int = min_segment_length
        self._once_fetch_size: int = once_fetch_size

        self._did_dispose: bool = False
        self._singleton_lock: Lock = Lock()
        self._singleton_phase: _SingletonPhase = _SingletonPhase.NOT_STARTED
        self._range_lock: Lock = Lock()
        self._did_cancel_range: bool = False
        self._range_downloader: RangeDownloader | None = None
        try:
            self._range_downloader = RangeDownloader(
                file_path=file_path,
                http_options=http_options,
                min_segment_length=min_segment_length,
                once_fetch_size=once_fetch_size,
                excepted_etag=excepted_etag,
            )
        except RangeDownloadFailedError:
            pass

    def pop_downloading_task(self) -> Callable[[], None] | None:
        if self._did_dispose:
            return None

        with self._range_lock:
            range_downloader = self._validate_range_downloader()
            if range_downloader is not None:
                segment = range_downloader.serial.pop_segment()
                if not segment:
                    return None
                return lambda: self._download_segment(
                    range_downloader=range_downloader,
                    segment=segment,
                )

        download_file = self._file_path.parent / chunk_name(self._file_path, 0)
        with self._singleton_lock:
            phase = self._singleton_phase
            if phase == _SingletonPhase.POPPED_TASK or phase == _SingletonPhase.COMPLETED:
                return None
            if phase == _SingletonPhase.FAILED:
                download_file.unlink(missing_ok=True)
            self._singleton_phase = _SingletonPhase.POPPED_TASK

        return lambda: self._download_file(download_file)

    def _download_segment(self, range_downloader: RangeDownloader, segment: Segment) -> None:
        try:
            if not self._did_dispose:
                range_downloader.download_segment(segment)

        except RangeDownloadFailedError as error:
            if not error.is_canceled_by:
                with self._range_lock:
                    range_downloader.serial.interrupt()
                    self._did_cancel_range = True
            raise error

        finally:
            segment.dispose()

    def _download_file(self, file_path: Path) -> None:
        try:
            resp = requests.Session().get(
                stream=True,
                url=self._http_options.url,
                headers=self._http_options.headers,
                cookies=self._http_options.cookies,
                timeout=self._http_options.timeout,
            )
            if resp.status_code in CAN_RETRY_STATUS_CODES:
                raise CanRetryError(f"HTTP {resp.status_code} - {resp.reason}")
            resp.raise_for_status()

            with open(file_path, "wb") as file:
                for chunk in resp.iter_content(self._once_fetch_size):
                    if len(chunk) == 0:
                        break
                    if self._did_dispose:
                        with self._singleton_lock:
                            self._singleton_phase = _SingletonPhase.FAILED
                        return
                    file.write(chunk)

            with self._singleton_lock:
                self._singleton_phase = _SingletonPhase.COMPLETED

        except Exception as error:
            with self._singleton_lock:
                self._singleton_phase = _SingletonPhase.FAILED
            if is_exception_can_retry(error):
                raise CanRetryError("Download file failed") from error
            raise error

    def try_complete(self) -> Path | None:
        chunk_paths: list[Path] = []
        with self._range_lock:
            range_downloader = self._validate_range_downloader()

        if range_downloader is not None:
            serial = range_downloader.serial
            if not serial.is_completed:
                return None
            for description in serial.snapshot():
                chunk_path = self._file_path.parent / chunk_name(self._file_path, description.offset)
                chunk_paths.append(chunk_path)
        else:
            with self._singleton_lock:
                if self._singleton_phase != _SingletonPhase.COMPLETED:
                    return None
            file_path = self._file_path.parent / chunk_name(self._file_path, 0)
            chunk_paths.append(file_path)

        clean_path(self._file_path)

        if len(chunk_paths) == 1:
            source_path = chunk_paths[0]
            shutil.move(source_path, self._file_path)
        else:
            try:
                with open(self._file_path, "wb") as output:
                    for chunk_path in chunk_paths:
                        with open(chunk_path, "rb") as input:
                            while True:
                                chunk = input.read(self._once_fetch_size)
                                if not chunk:
                                    break
                                output.write(chunk)

            except Exception as err:
                self._file_path.unlink(missing_ok=True)
                raise err

            for chunk_path in chunk_paths:
                chunk_path.unlink(missing_ok=True)
        return self._file_path

    def dispose(self) -> None:
        if self._did_dispose:
            return
        self._did_dispose = True

        with self._range_lock:
            range_downloader = self._range_downloader
            if range_downloader is not None:
                range_downloader.serial.dispose()
                self._range_downloader = None
                self._did_cancel_range = False

    def _validate_range_downloader(self) -> RangeDownloader | None:
        range_downloader = self._range_downloader
        if self._did_cancel_range and range_downloader is not None:
            range_downloader.serial.dispose()
            range_downloader = None
            self._range_downloader = None

        return range_downloader



from typing import TypeVar

T = TypeVar("T")


def list_safe_remove(lst: list[T], item: T) -> T | None:
    index = -1
    for i, elem in enumerate(lst):
        if elem is item:
            index = i
            break
    if index >= 0:
        return lst.pop(index)



import glob
from collections.abc import Mapping
from pathlib import Path

import requests

from ..common import HTTPOptions
from ..retry import CAN_RETRY_STATUS_CODES, Retry, is_exception_can_retry
from .common import DOWNLOADING_SUFFIX, chunk_name
from .errors import CanRetryError, RangeDownloadFailedError
from .segment import Segment, SegmentDescription, Serial
from .utils import clean_path
from .value_signal import ValueSignal


# thread safe
class RangeDownloader:
    def __init__(
        self,
        file_path: Path,
        http_options: HTTPOptions,
        min_segment_length: int,
        once_fetch_size: int,
        excepted_etag: str | None = None,
    ) -> None:
        self._file_path: Path = file_path
        self._http_options: HTTPOptions = http_options
        self._min_segment_length: int = min_segment_length
        self._once_fetch_size: int = once_fetch_size
        self._support_range_signal: ValueSignal[bool] = ValueSignal()

        content_length, etag, range_useable = self._fetch_meta(http_options.retry)
        if content_length is None:
            raise RangeDownloadFailedError("Content-Length header is missing in response")
        if not range_useable:
            raise RangeDownloadFailedError("Server does not support Range requests")

        descriptions = self._create_segment_descriptions(
            content_length=content_length,
            etag=etag,
            excepted_etag=excepted_etag,
        )
        self._serial: Serial = Serial(
            length=content_length,
            min_segment_length=self._min_segment_length,
            descriptions=descriptions,
        )

    def _fetch_meta(self, retry: Retry):
        resp = retry.request(
            request=lambda: requests.head(
                url=self._http_options.url,
                headers=self._http_options.headers,
                cookies=self._http_options.cookies,
                timeout=self._http_options.timeout,
            ),
        )
        content_length = resp.headers.get("Content-Length")
        etag = resp.headers.get("ETag")
        range_useable = resp.headers.get("Accept-Ranges") == "bytes"

        if content_length is not None:
            content_length = int(content_length)
        return content_length, etag, range_useable

    def _create_segment_descriptions(
        self, content_length: int, etag: str | None, excepted_etag: str | None
    ) -> list[SegmentDescription] | None:
        descriptions: list[SegmentDescription] | None = None
        offsets: list[int] | None = None

        if etag is not None and excepted_etag is not None and excepted_etag != etag:
            for offset in self._search_offsets(content_length):
                clean_path(self._file_path.parent / chunk_name(self._file_path, offset))
        elif self._file_path.parent.exists():
            assert self._file_path.parent.is_dir(), f"{self._file_path.parent} is not a directory"
            offsets = list(self._search_offsets(content_length))
            offsets.sort()
        else:
            self._file_path.parent.mkdir(parents=True, exist_ok=True)

        if offsets:
            descriptions = []
            for i, offset in enumerate(offsets):
                length: int
                if i == len(offsets) - 1:
                    length = content_length - offset
                else:
                    length = offsets[i + 1] - offset

                chunk_path = self._file_path.parent / chunk_name(self._file_path, offset)
                chunk_size = chunk_path.stat().st_size
                trim_size = chunk_size - length
                if trim_size > 0:
                    self._trim_file_tail(chunk_path, trim_size)

                descriptions.append(
                    SegmentDescription(
                        offset=offset,
                        length=length,
                        completed_length=chunk_size,
                    )
                )

        return descriptions

    def _search_offsets(self, length: int):
        wanna_tail = f"{self._file_path.suffix[1:]}{DOWNLOADING_SUFFIX}"
        file_stem = glob.escape(self._file_path.stem)  # maybe include "*" signs

        for matched_path in self._file_path.parent.glob(f"{file_stem}*"):
            matched_tail = matched_path.name[len(self._file_path.stem) :]
            if matched_tail == DOWNLOADING_SUFFIX:
                yield 0
            else:
                parts = matched_tail.split(".", maxsplit=2)
                if len(parts) == 3 and parts[0] == "":
                    _, str_offset, tail = parts
                    if tail == wanna_tail:
                        offset: int = -1
                        try:
                            offset = int(str_offset)
                        except ValueError:
                            pass
                        if 0 < offset < length:
                            yield offset

    @property
    def serial(self) -> Serial:
        return self._serial

    def download_segment(self, segment: Segment) -> None:
        chunk_path = self._file_path.parent / chunk_name(self._file_path, segment.offset)
        chunk_size = 0
        if chunk_path.exists():
            chunk_size = chunk_path.stat().st_size

        if chunk_size >= segment.length:
            trim_size = chunk_size - segment.length
            if trim_size > 0:
                self._trim_file_tail(chunk_path, trim_size)
            segment.complete()
            return

        # 服务器可能变卦，在 meta 中声明支持 Range，但真正 fetch 时又不支持了
        # 只有发起一次真正的 GET 请求，然后从 Response Head 中读取信息才能知道到底值不支持
        # 因此先让第一个 Request 发起请求，并同时阻塞其他任务，直到第一个请求的 HEAD 部分返回再进行下一步操作
        with self._support_range_signal.context() as support_range_context:
            if support_range_context.value == False:  # noqa: E712
                raise RangeDownloadFailedError(
                    message="Task is canceled because server rejects Range request",
                    is_canceled_by=True,
                )
            try:
                response = self._create_download_segment_response(segment)
            except RangeDownloadFailedError as error:
                if not error.is_canceled_by:
                    support_range_context.update_value(False)
                raise error

            support_range_context.update_value(True)
            self._download_segment_into_file(response, chunk_path, segment)

    def _trim_file_tail(self, file_path: Path, bytes: int) -> None:
        with open(file_path, "rb+") as file:
            file.seek(0, 2)
            size = file.tell()
            if size >= bytes:
                file.truncate(size - bytes)

    def _create_download_segment_response(self, segment: Segment):
        download_start = segment.offset + segment.completed_length
        download_end = segment.offset + segment.length - 1
        download_length = download_end - download_start + 1

        headers: Mapping[str, str | bytes | None] = {}
        if self._http_options.headers:
            headers.update(self._http_options.headers)

        headers["Range"] = f"bytes={download_start}-{download_end}"
        response = requests.Session().get(
            stream=True,
            url=self._http_options.url,
            headers=headers,
            cookies=self._http_options.cookies,
            timeout=self._http_options.timeout,
        )
        if response.status_code in CAN_RETRY_STATUS_CODES:
            raise CanRetryError(f"HTTP {response.status_code} - {response.reason}")
        if response.status_code == 416:
            raise RangeDownloadFailedError("Server rejects Range request")
        response.raise_for_status()

        if response.status_code == 206:
            content_range = response.headers.get("Content-Range")
            content_length = response.headers.get("Content-Length")

            if content_range != f"bytes {download_start}-{download_end}/{self._serial.length}":
                raise RangeDownloadFailedError(f"Unexpected Content-Range: {content_range}")
            if content_length != f"{download_length}":
                raise RangeDownloadFailedError(f"Unexpected Content-Length: {content_length}")

        elif response.status_code == 200:
            if download_start != 0:
                raise RangeDownloadFailedError("Server returns 200 OK but Range request was made")
            content_length = response.headers.get("Content-Length")
            if content_length != str(download_length):
                raise RangeDownloadFailedError(f"Server returns 200 OK but Unexpected Content-Length: {content_length}")

        else:
            raise RangeDownloadFailedError(f"Server rejects Range request with status code {response.status_code}")

        return response

    def _download_segment_into_file(self, response: requests.Response, chunk_path: Path, segment: Segment) -> None:
        with open(chunk_path, "ab") as file:
            try:
                for chunk in response.iter_content(self._once_fetch_size):
                    chunk_size = len(chunk)
                    writable_size = segment.lock_next_length(chunk_size)
                    if writable_size == 0:
                        break
                    if writable_size == chunk_size:
                        file.write(chunk)
                    else:
                        file.write(chunk[:writable_size])
                    segment.submit(writable_size)

            except Exception as error:
                if is_exception_can_retry(error):
                    raise CanRetryError("Download one segment of the file failed") from error
                else:
                    raise error

        if not segment.is_completed:
            raise CanRetryError("Connection closed before completing segment")



from pathlib import Path

DOWNLOADING_SUFFIX = ".downloading"


def chunk_name(path: Path, offset: int) -> str:
    if offset == 0:
        return f"{path.name}{DOWNLOADING_SUFFIX}"
    else:
        return f"{path.stem}.{offset}{path.suffix}{DOWNLOADING_SUFFIX}"



from dataclasses import dataclass
from os import PathLike
from typing import Callable, Mapping, MutableMapping


@dataclass
class Task:
    url: str | Callable[[], str]
    file: str | PathLike
    headers: Mapping[str, str | bytes | None] | None = None
    cookies: MutableMapping[str, str] | None = None

    def get_url(self) -> str:
        if callable(self.url):
            self.url = self.url()
        return self.url


class TaskError(Exception):
    def __init__(self, task: Task, case_error: Exception) -> None:
        super().__init__(f"URL download failed: {task.url}")
        self._task: Task = task
        self._case_error: Exception = case_error

    @property
    def task(self) -> Task:
        return self._task

    @property
    def case_error(self) -> Exception:
        return self._case_error


class RetryError(TaskError):
    pass


class TooManyRetriesError(TaskError):
    pass


class FileDownloadError(TaskError):
    pass



from pathlib import Path
from shutil import rmtree


def clean_path(path: Path) -> None:
    if not path.exists():
        return
    if path.is_file():
        path.unlink()
    else:
        rmtree(path)



from threading import Lock, Semaphore
from typing import Callable, Generic, TypeVar

D = TypeVar("D")


class SafePassageClosed(Exception):
    pass


class SafePassage(Generic[D]):
    def __init__(self) -> None:
        super().__init__()
        self._main_lock: Lock = Lock()
        self._producer_lock: Lock = Lock()
        self._consumer_lock: Lock = Lock()
        self._sender_semaphore: Semaphore = Semaphore(0)
        self._consumer_can_receive_semaphore: Semaphore = Semaphore(0)
        self._consumer_did_receive_semaphore: Semaphore = Semaphore(0)
        self._buffer: tuple[D] | Exception | None = None
        self._rejected_error: Exception | None = None
        self._did_close: bool = False

    @property
    def did_close(self) -> bool:
        with self._main_lock:
            return self._did_close

    def send(self, data: D) -> None:
        self._run_send((data,))

    def build(self, builder: Callable[[], D]) -> None:
        self._run_send(builder)

    def _run_send(self, target: tuple[D] | Callable[[], D]) -> None:
        with self._producer_lock:
            with self._main_lock:
                self._assert_not_closed()

            self._sender_semaphore.acquire()

            with self._main_lock:
                self._assert_not_closed()
                if self._rejected_error is not None:
                    error = self._rejected_error
                    self._rejected_error = None
                    self._consumer_can_receive_semaphore.release()
                    raise error

                if isinstance(target, tuple):
                    self._buffer = target
                else:
                    try:
                        self._buffer = (target(),)
                    except Exception as error:
                        self._buffer = error
                self._consumer_can_receive_semaphore.release()

            # 此锁用于避免 consumer 没有收到之前， producer 就 close 了。
            self._consumer_did_receive_semaphore.acquire()

    def close(self) -> None:
        with self._producer_lock, self._main_lock:
            if not self._did_close:
                self._did_close = True
                self._sender_semaphore.release()
                self._consumer_can_receive_semaphore.release()

    def receive(self) -> D:
        with self._consumer_lock:
            with self._main_lock:
                self._assert_not_closed()
                self._sender_semaphore.release()

            self._consumer_can_receive_semaphore.acquire()
            try:
                with self._main_lock:
                    self._assert_not_closed()
                    buffer = self._buffer
                    self._buffer = None
                    assert buffer is not None
                    if isinstance(buffer, Exception):
                        raise buffer
                    else:
                        return buffer[0]

            finally:
                self._consumer_did_receive_semaphore.release()

    def reject(self, error: Exception) -> None:
        with self._consumer_lock:
            with self._main_lock:
                self._assert_not_closed()
                self._rejected_error = error
                self._sender_semaphore.release()
            self._consumer_can_receive_semaphore.acquire()

    def _assert_not_closed(self) -> None:
        if self._did_close:
            raise SafePassageClosed()



class InterruptionError(Exception):
    def __init__(self) -> None:
        super().__init__("Interrupted")


class CanRetryError(Exception):
    pass


class RangeDownloadFailedError(CanRetryError):
    def __init__(self, message: str, is_canceled_by: bool = False) -> None:
        super().__init__(message)
        self._is_canceled_by: bool = is_canceled_by

    @property
    def is_canceled_by(self) -> bool:
        return self._is_canceled_by



import http.client
import socket
import ssl
from time import sleep
from typing import Callable

import requests
import urllib3.exceptions
from requests import Response

CAN_RETRY_STATUS_CODES = (408, 429, 502, 503, 504)

_CAN_RETRY_EXCEPTIONS = (
    # requests 层异常
    requests.exceptions.ConnectionError,
    requests.exceptions.Timeout,
    requests.exceptions.ProxyError,
    requests.exceptions.ChunkedEncodingError,
    requests.exceptions.ReadTimeout,
    requests.exceptions.ConnectTimeout,
    requests.exceptions.TooManyRedirects,
    requests.exceptions.ContentDecodingError,  # 内容解码错误（gzip等）
    requests.exceptions.HTTPError,  # HTTP 错误状态码
    # urllib3 层异常（重要！可能直接抛出）
    urllib3.exceptions.ConnectionError,
    urllib3.exceptions.TimeoutError,
    urllib3.exceptions.ReadTimeoutError,
    urllib3.exceptions.ConnectTimeoutError,
    urllib3.exceptions.NewConnectionError,
    urllib3.exceptions.ProxyError,
    urllib3.exceptions.ProtocolError,
    urllib3.exceptions.MaxRetryError,
    urllib3.exceptions.ClosedPoolError,
    urllib3.exceptions.EmptyPoolError,
    urllib3.exceptions.HostChangedError,  # 主机变更错误
    urllib3.exceptions.DecodeError,  # 内容解码错误
    urllib3.exceptions.ResponseError,  # 响应错误
    #
    # SSL异常处理参考文档：
    # - Python SSL官方文档: https://docs.python.org/3/library/ssl.html
    # - SSL异常详解: https://medium.com/@obaff/possible-causes-and-solutions-for-sslerror-ssleoferror-8-eof-occurred-in-violation-of-protocol-9ad28ff56d56
    # - urllib3 SSLEOFError行为变化: https://github.com/urllib3/urllib3/issues/3100
    # - requests SSLZeroReturnError问题: https://github.com/psf/requests/issues/6429
    #
    # SSL异常层次结构：
    #     OSError
    #      └── ssl.SSLError (基类 - 不应直接捕获，应捕获具体子类)
    #           ├── ssl.SSLZeroReturnError - SSL连接已干净关闭 (可重试)
    #           ├── ssl.SSLWantReadError - 非阻塞socket需要读 (可重试，但requests中不应出现)
    #           ├── ssl.SSLWantWriteError - 非阻塞socket需要写 (可重试，但requests中不应出现)
    #           ├── ssl.SSLSyscallError - 系统错误 (可重试)
    #           ├── ssl.SSLEOFError - SSL连接被意外关闭 (可重试)
    #           └── ssl.SSLCertVerificationError - 证书验证失败 (不可重试!)
    #
    # 标准库 SSL 层异常（最底层，最关键！）
    # 注意：不包含 ssl.SSLError 基类，避免捕获不可重试的 ssl.SSLCertVerificationError
    # 注意：不包含 requests.exceptions.SSLError 和 urllib3.exceptions.SSLError
    #       因为它们是包装类，会包装所有底层SSL异常（包括不可重试的证书验证失败）
    #       正确做法：只捕获具体的、可重试的 ssl.* 子类，通过异常链遍历来识别
    ssl.SSLEOFError,  # SSL连接被意外关闭 (UNEXPECTED_EOF_WHILE_READING)
    ssl.SSLZeroReturnError,  # SSL连接已干净关闭（连接池复用时常见）
    ssl.SSLSyscallError,  # SSL系统调用错误（底层socket/系统问题）
    ssl.SSLWantReadError,  # 非阻塞socket需要读（理论上requests中不应出现）
    ssl.SSLWantWriteError,  # 非阻塞socket需要写（理论上requests中不应出现）
    # 标准库 socket 层异常
    socket.timeout,  # 套接字超时
    socket.error,  # 套接字错误
    # Python 内置异常
    ConnectionError,  # Python 3.3+ 连接错误
    TimeoutError,  # Python 3.3+ 超时错误
    OSError,  # 底层 I/O 错误 (包含连接重置)
    IOError,  # I/O 错误（兼容性考虑）
    # HTTP 协议层异常（iter_content 阶段常见）
    http.client.HTTPException,  # HTTP 协议通用异常
    http.client.IncompleteRead,  # 读取不完整（网络中断）
    http.client.BadStatusLine,  # 状态行错误
    http.client.ResponseNotReady,  # 响应未准备好
    http.client.RemoteDisconnected,  # 远程连接被关闭
)

_CAN_RETRY_KEYWORDS = (
    "eof occurred in violation of protocol",
    "connection reset",
    "connection refused",
    "timeout",
)


def is_exception_can_retry(error: Exception) -> bool:
    for current_exc in _walk_exception_chain(error, max_depth=12):
        for retry_class in _CAN_RETRY_EXCEPTIONS:
            if isinstance(current_exc, retry_class):
                return True

    error_msg = str(error).lower()
    for keyword in _CAN_RETRY_KEYWORDS:
        if keyword in error_msg:
            return True

    return False


def _walk_exception_chain(error: Exception, max_depth: int):
    current_exc = error
    depth = 0

    yield current_exc

    while hasattr(current_exc, "__cause__") and current_exc.__cause__ is not None and depth < max_depth:
        current_exc = current_exc.__cause__
        depth += 1
        yield current_exc


class Retry:
    def __init__(
        self,
        retry_times: int,
        retry_sleep: float,
    ) -> None:
        self._retry_times: int = retry_times
        self._retry_sleep: float = retry_sleep

    def request(self, request: Callable[[], Response]) -> Response:
        last_response: Response | None = None
        last_error: Exception | None = None
        for i in range(self._retry_times + 1):
            try:
                last_response = request()
                if last_response.ok:
                    return last_response
                if last_response.status_code not in CAN_RETRY_STATUS_CODES:
                    break
            except Exception as error:
                if is_exception_can_retry(error):
                    last_error = error
                else:
                    raise error

            if i < self._retry_times:
                sleep(self._retry_sleep)

        if last_error is not None:
            raise last_error

        if last_response is not None:
            last_response.raise_for_status()

        raise RuntimeError("request failed")



from threading import Lock, Thread
from typing import Callable, Iterable, Iterator

from .common import Retry
from .files_group import FilesGroup
from .safe_passage import SafePassage, SafePassageClosed
from .type import RetryError, Task, TaskError


def download(
    tasks_iter: Task | Iterator[Task],
    window_width: int = 7,
    threads_count: int = 1,
    failure_ladder: int | Iterable[int] = 5,
    min_segment_length: int = 1024 * 1024,  # 1 MiB
    once_fetch_size: int = 1024 * 16,
    timeout: float = 30.0,
    retry_times: int = 0,
    retry_sleep: float = 0.0,
    override_existing_files: bool = False,
    on_task_completed: Callable[[Task], None] | None = None,
    on_task_skipped: Callable[[Task], None] | None = None,
    on_task_failed: Callable[[TaskError], None] | None = None,
    on_task_failed_with_retry_error: Callable[[RetryError], None] | None = None,
) -> None:

    assert threads_count > 0, "threads_count must be greater than zero"

    if isinstance(tasks_iter, Task):
        tasks_iter = iter([tasks_iter])

    if isinstance(failure_ladder, int):
        failure_ladder = (failure_ladder,)

    group = FilesGroup(
        tasks_iter=tasks_iter,
        window_width=window_width,
        failure_ladder=failure_ladder,
        min_segment_length=min_segment_length,
        once_fetch_size=once_fetch_size,
        skip_existing=not override_existing_files,
        timeout=timeout,
        on_task_completed=on_task_completed,
        on_task_skipped=on_task_skipped,
        on_task_failed=on_task_failed,
        on_task_failed_with_retry_error=on_task_failed_with_retry_error,
        retry=Retry(
            retry_times=retry_times,
            retry_sleep=retry_sleep,
        ),
    )
    threads: list[Thread] = []
    invoker = _Invoker(
        group=group,
        threads_count=threads_count,
        on_task_failed=on_task_failed,
    )
    for _ in range(threads_count):
        thread = Thread(target=invoker.run_job)
        thread.start()
        threads.append(thread)

    try:
        invoker.run_main()

    finally:
        for thread in threads:
            thread.join()


class _Invoker:
    def __init__(
        self,
        group: FilesGroup,
        threads_count: int,
        on_task_failed: Callable[[TaskError], None] | None,
    ) -> None:

        self._group: FilesGroup = group
        self._threads_count: int = threads_count
        self._on_task_failed: Callable[[TaskError], None] | None = on_task_failed
        self._passage: SafePassage[Callable[[], None] | None] = SafePassage()
        self._failure_lock: Lock = Lock()
        self._failure_mark: bool = False

    def run_main(self) -> None:
        found_no_more = False
        found_error: Exception | None = None

        def build_executor() -> Callable[[], None] | None:
            nonlocal found_no_more, found_error
            found_no_more = False
            found_error = None
            try:
                executor = self._group.pop_downloading_executor()
                found_no_more = executor is None
                return executor
            except Exception as error:
                found_error = error
                return None

        try:
            while True:
                while True:
                    self._passage.build(build_executor)
                    if found_error is not None:
                        if self._on_task_failed and isinstance(found_error, TaskError):
                            self._on_task_failed(found_error)
                        else:
                            raise found_error
                    if found_no_more:
                        break

                if self._group.is_empty:
                    break
                self._group.wait_windows_update()

        except Exception as error:
            self._group.dispose()
            raise error

        finally:
            self._passage.close()

    def run_job(self):
        try:
            while True:
                executor = self._passage.receive()
                if executor:
                    executor()
                    self._group.raise_if_failure()

        except SafePassageClosed:
            pass

        except Exception as error:
            with self._failure_lock:
                if not self._failure_mark:
                    self._failure_mark = True
                    self._passage.reject(error)



from dataclasses import dataclass
from typing import Mapping, MutableMapping

from .retry import Retry


@dataclass
class HTTPOptions:
    url: str
    timeout: float
    retry: Retry
    headers: Mapping[str, str | bytes | None] | None
    cookies: MutableMapping[str, str] | None
